<?php

declare(strict_types=1);

/*
 * This file is part of Contao.
 *
 * (c) Leo Feyer
 *
 * @license LGPL-3.0-or-later
 */

namespace Contao\CoreBundle\Twig\Extension;

use Twig\Environment;
use Twig\Node\DeprecatedNode;
use Twig\Node\Expression\ConstantExpression;
use Twig\Node\Node;
use Twig\Node\PrintNode;
use Twig\NodeVisitor\AbstractNodeVisitor;

/**
 * @internal
 */
class DeprecationsNodeVisitor extends AbstractNodeVisitor
{
    public function getPriority(): int
    {
        return 10;
    }

    protected function doEnterNode(Node $node, Environment $env): Node
    {
        return $node;
    }

    protected function doLeaveNode(Node $node, Environment $env): Node
    {
        return $this->handleDeprecatedInsertTagUsage($node);
    }

    /**
     * Discourage the use of insert tags as strings like "{{ '{{link_url::9}}' }}"
     * instead of directly evaluating them via the "insert_tag" function.
     */
    private function handleDeprecatedInsertTagUsage(Node $node): Node
    {
        $insertTagMisusePattern = '/{{([^}]+)}}/';

        if (
            !$node instanceof PrintNode
            || !($expression = $node->getNode('expr')) instanceof ConstantExpression
            || 1 !== preg_match($insertTagMisusePattern, (string) $expression->getAttribute('value'), $matches)
        ) {
            return $node;
        }

        $suggestedTransformation = sprintf('"{{ \'{{%1$s}}\' }}" -> "{{ insert_tag(\'%1$s\') }}".', $matches[1]);

        $message = 'You should not rely on insert tags being replaced in the rendered HTML. '
            .'This behavior will gradually be phased out in Contao 5 and will no longer work in Contao 6.0. '
            .'Explicitly replace insert tags with the "insert_tag" function instead: '.$suggestedTransformation;

        return $this->addDeprecation($node, $message);
    }

    private function addDeprecation(Node $node, string $message): Node
    {
        $line = $node->getTemplateLine();

        $deprecatedNode = new DeprecatedNode(
            new ConstantExpression("Since contao/core-bundle 4.13: $message", $line),
            $line,
            $node->getNodeTag()
        );

        // Set the source context, so that the template name can be inserted
        // when compiling the DeprecatedNode.
        $deprecatedNode->setSourceContext($node->getSourceContext());

        return new Node([$node, $deprecatedNode]);
    }
}
